#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import collections

import octobot_commons.logging as logging
import octobot_commons.enums as commons_enums
import octobot_commons.tree as commons_tree

import octobot_trading.personal_data.positions.position_factory as position_factory
import octobot_trading.util as util
import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.exchange_channel as exchange_channel


class PositionsManager(util.Initializable):
    POSITION_ID_SEPARATOR = "_"

    def __init__(self, trader):
        super().__init__()
        self.logger = logging.get_logger(self.__class__.__name__)
        self.trader = trader
        self.positions = collections.OrderedDict()
        self.logged_unsupported_positions = set()

    async def initialize_impl(self):
        self._reset_positions()

    def get_symbol_position(self, symbol, side):
        """
        Returns or create the symbol position instance
        :param symbol: the position symbol
        :param side: the position side
        :return: the existing position or the newly created position
        """
        return self._get_or_create_position(symbol=symbol, side=side)

    def get_order_position(self, order, contract=None):
        """
        Returns the position that matches the order
        :param order: the order
        :param contract: the symbol contract (optional)
        :return: the existing position or the newly created position that matches the order
        """
        future_contract = contract if contract is not None \
            else self.trader.exchange_manager.exchange.get_pair_future_contract(order.symbol)
        return self.get_symbol_position(symbol=order.symbol,
                                        side=None if future_contract.is_one_way_position_mode()
                                        else order.get_position_side(future_contract))

    def get_symbol_positions(self, symbol=None):
        """
        Returns symbol positions if exist
        :param symbol: the position symbol
        :return: the symbol positions
        """
        if symbol is None:
            return list(self.positions.values())
        return self._get_symbol_positions(symbol)

    async def upsert_position(self, symbol: str, side, raw_position: dict) -> bool:
        """
        Create or update a position from a raw dictionary
        :param symbol: the position symbol
        :param side: the position side
        :param raw_position: the position raw dictionary
        :return: True when the creation or the update succeeded
        """
        self._ensure_support(raw_position)
        position_id = self._generate_position_id(symbol=symbol, side=side)
        if position_id not in self.positions:
            new_position = position_factory.create_position_instance_from_raw(self.trader, raw_position=raw_position)
            new_position.position_id = position_id
            return await self._finalize_position_creation(new_position, is_from_exchange_data=True)

        return self.positions[position_id].update_from_raw(raw_position)

    def _ensure_support(self, raw_position):
        if (
            raw_position.get(enums.ExchangeConstantsPositionColumns.POSITION_MODE.value, enums.PositionMode.ONE_WAY)
            is not enums.PositionMode.ONE_WAY
            and raw_position[enums.ExchangeConstantsPositionColumns.SYMBOL.value]
            not in self.logged_unsupported_positions
        ):
            # TODO important error to display
            self.logger.error(
                f"{raw_position[enums.ExchangeConstantsPositionColumns.SYMBOL.value]} position is in "
                f"{raw_position[enums.ExchangeConstantsPositionColumns.POSITION_MODE.value].name} mode. "
                f"This mode is not supported and will create unexpected behaviors in OctoBot. Please switch "
                f"to {enums.PositionMode.ONE_WAY.name} mode on {self.trader.exchange_manager.exchange_name}."
            )
            self.logged_unsupported_positions.add(raw_position[enums.ExchangeConstantsPositionColumns.SYMBOL.value])

    def set_initialized_event(self, symbol):
        commons_tree.EventProvider.instance().trigger_event(
            self.trader.exchange_manager.bot_id, commons_tree.get_exchange_path(
                self.trader.exchange_manager.exchange_name,
                commons_enums.InitializationEventExchangeTopics.POSITIONS.value,
                symbol=symbol
            )
        )

    async def recreate_position(self, position) -> bool:
        """
        Recreate position from an existing position instance
        :param position: the position instance to recreate
        :return: True when the recreation succeeded
        """
        new_position = position_factory.create_position_instance_from_raw(self.trader, raw_position=position.to_dict())
        position.clear()
        position.position_id = self._generate_position_id(symbol=position.symbol, side=position.side)
        return await self._finalize_position_creation(new_position)

    async def handle_position_update_from_order(self, order, require_exchange_update: bool) -> bool:
        """
        Handle a position update from an order update
        :param order: the order
        :param require_exchange_update: when True, will sync with exchange position, otherwise will predict the
        position changes using order data (as in trading simulator)
        :return: True if the position was updated
        """
        if self.trader.is_enabled:
            # portfolio might be updated when refreshing the position
            async with self.trader.exchange_manager.exchange_personal_data.portfolio_manager.portfolio_history_update():
                if self.trader.simulate or not require_exchange_update:
                    # update simulated positions
                    return self._refresh_simulated_position_from_order(order)
                if require_exchange_update and order.is_filled():
                    # on real trading when orders is filled: reload positions to ensure positions sync
                    try:
                        await self.refresh_real_trader_position(self.get_order_position(order))
                        return True
                    except Exception as err:
                        self.logger.exception(
                            err, True, f"Error while refreshing real trader {order.symbol} position: {err}"
                        )
        return False

    def _refresh_simulated_position_from_order(self, order):
        if order.is_filled():
            # Don't update if order filled quantity is null
            if order.filled_quantity == 0:
                return False

            position_instance = order.exchange_manager.exchange_personal_data.positions_manager.get_order_position(
                order, contract=order.exchange_manager.exchange.get_pair_future_contract(order.symbol))
            try:
                position_instance.update_from_order(order)
                return True
            except errors.PortfolioNegativeValueError as portfolio_negative_value_error:
                self.logger.exception(portfolio_negative_value_error, True,
                                      f"Failed to update portfolio via position : {portfolio_negative_value_error} "
                                      f"for order {order.to_dict()}")
        return False

    async def refresh_real_trader_position(self, position, force_job_execution=False):
        """
        :param position: the position instance to refresh
        :param force_job_execution: force_job_execution
        Call POSITIONS_CHANNEL producer to refresh real trader position
        """
        await exchange_channel.get_chan(
            constants.POSITIONS_CHANNEL, self.trader.exchange_manager.id
        ).get_internal_producer().update_position_from_exchange(
            position, wait_for_refresh=True, force_job_execution=force_job_execution
        )

    def upsert_position_instance(self, position) -> bool:
        """
        Save an existing position instance to positions list
        :param position: the position instance
        :return: True when the operation succeeded
        """
        if position.position_id not in self.positions:
            self.positions[position.position_id] = position
            return True
        return False

    def clear(self):
        """
        Clear all positions and the position OrderedDict
        """
        for position in self.positions.values():
            position.clear()
        self._reset_positions()

    # private
    def _generate_position_id(self, symbol, side, expiration_time=None):
        """
        Generate a position ID for one way and hedge position modes
        :param symbol: the position symbol
        :param side: the position side
        :param expiration_time: the symbol expiration timestamp
        :return: the computed position id
        """
        return f"{symbol}" \
               f"{'' if expiration_time is None else self.POSITION_ID_SEPARATOR + str(expiration_time)}" \
               f"{'' if side is enums.PositionSide.BOTH or side is None else self.POSITION_ID_SEPARATOR + side.value}"

    async def _finalize_position_creation(self, new_position, is_from_exchange_data=False) -> bool:
        """
        Ends a position creation process
        :param new_position: the new position instance
        :param is_from_exchange_data: True when the exchange creation comes from exchange data
        :return: True when the process succeeded
        """
        self.positions[new_position.position_id] = new_position
        await new_position.initialize(is_from_exchange_data=is_from_exchange_data)
        return True

    def _create_symbol_position(self, symbol, position_id):
        """
        Creates a position when it doesn't exist for the specified symbol
        :param symbol: the new position symbol
        :param side: the new position id
        :return: the new symbol position instance
        """
        new_position = position_factory.create_symbol_position(self.trader, symbol)
        new_position.position_id = position_id
        self.positions[position_id] = new_position
        return new_position

    def _get_or_create_position(self, symbol, side):
        """
        Get or create position by symbol and side
        :param symbol: the expected position symbol
        :param side: the expected position side
        :return: the matching position
        """
        expected_position_id = self._generate_position_id(symbol=symbol, side=side)
        try:
            return self.positions[expected_position_id]
        except KeyError:
            self.positions[expected_position_id] = self._create_symbol_position(symbol, expected_position_id)
        return self.positions[expected_position_id]

    def _get_symbol_positions(self, symbol):
        """
        Get symbol positions in each side
        :param symbol: the position symbol
        :return: existing symbol positions list
        """
        positions = []
        for side in [enums.PositionSide.BOTH, enums.PositionSide.SHORT, enums.PositionSide.LONG]:
            position_id = self._generate_position_id(symbol=symbol, side=side)
            try:
                positions.append(self.positions[position_id])
            except KeyError:
                pass
        return positions

    def _reset_positions(self):
        """
        Clear all position references
        """
        self.positions = collections.OrderedDict()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import octobot_commons.enums as commons_enums
import octobot_commons.authentication as authentication
import octobot_commons.databases as commons_databases
import octobot_commons.tree as commons_tree
import octobot_commons.constants as commons_constants

import octobot_trading.storage.abstract_storage as abstract_storage
import octobot_trading.storage.util as storage_util
import octobot_trading.personal_data.portfolios.history as portfolio_history
import octobot_trading.exchanges as exchanges


class PortfolioStorage(abstract_storage.AbstractStorage):
    IS_LIVE_CONSUMER = False
    IS_HISTORICAL = True
    PRICE_INIT_TIMEOUT = 30
    HISTORY_TABLE = commons_enums.RunDatabases.HISTORICAL_PORTFOLIO_VALUE.value

    @abstract_storage.AbstractStorage.hard_reset_and_retry_if_necessary
    async def store_history(self, reset=False):
        if not self.enabled:
            return
        portfolio_db = self.get_db()
        hist_portfolio_values_manager = self.exchange_manager.exchange_personal_data.\
            portfolio_manager.historical_portfolio_value_manager
        metadata = hist_portfolio_values_manager.get_metadata()
        # replace the whole table to ensure consistency
        history = hist_portfolio_values_manager.get_dict_historical_values()
        existing_history = await portfolio_db.all(self.HISTORY_TABLE)
        self._to_update_auth_data_ids_buffer.update(
            (
                history_val[portfolio_history.HistoricalAssetValue.TIMESTAMP_KEY]
                for history_val in history
                if history_val not in existing_history
            )
        )
        await portfolio_db.upsert(commons_enums.RunDatabases.METADATA.value, metadata, None, uuid=1)
        await portfolio_db.replace_all(
            self.HISTORY_TABLE,
            history,
            cache=False
        )
        await self.trigger_debounced_flush()
        await self.trigger_debounced_update_auth_data(reset)

    async def _update_auth_data(self, reset):
        authenticator = authentication.Authenticator.instance()
        if not authenticator.is_initialized():
            return
        current_value = 0
        initial_value_by_timestamp = {}
        reference_market = None
        ending_portfolio = {}
        price_by_asset = {}
        historical_value_by_timestamp = {}
        for exchange_manager in exchanges.Exchanges.instance().get_exchanges_managers_with_same_matrix_id(
            self.exchange_manager
        ):
            if not exchange_manager.exchange_personal_data.portfolio_manager:
                # exchange has no portfolio (probably not trading)
                continue
            hist_portfolio_values_manager = exchange_manager.exchange_personal_data. \
                portfolio_manager.historical_portfolio_value_manager
            if initializing_prices := hist_portfolio_values_manager.portfolio_manager.portfolio_value_holder.\
                value_converter.initializing_symbol_prices_pairs:
                for symbol in initializing_prices:
                    await commons_tree.EventProvider.instance().wait_for_event(
                        self.exchange_manager.bot_id,
                        commons_tree.get_exchange_path(
                            self.exchange_manager.exchange_name,
                            commons_enums.InitializationEventExchangeTopics.PRICE.value,
                            symbol=symbol,
                        ),
                        self.PRICE_INIT_TIMEOUT
                    )
            reference_market = hist_portfolio_values_manager.portfolio_manager.reference_market
            full_history = hist_portfolio_values_manager.get_dict_historical_values()
            if full_history:
                latest_value = full_history[-1]
                first_value = full_history[0]
            else:
                # use current value
                latest_value = {
                    portfolio_history.HistoricalAssetValue.VALUES_KEY: {
                        reference_market: float(
                            hist_portfolio_values_manager.portfolio_manager.portfolio_value_holder.portfolio_current_value
                        )
                    },
                    portfolio_history.HistoricalAssetValue.TIMESTAMP_KEY:
                        self.exchange_manager.exchange.get_exchange_current_time(),
                }
                first_value = latest_value
            if not (latest_value and first_value and hist_portfolio_values_manager.ending_portfolio):
                continue
            price_by_asset.update(
                hist_portfolio_values_manager.portfolio_manager.portfolio_value_holder.current_crypto_currencies_values
            )
            if hist_portfolio_values_manager.ending_portfolio:
                for asset, value in hist_portfolio_values_manager.ending_portfolio.items():
                    if asset not in ending_portfolio:
                        ending_portfolio[asset] = {
                            commons_constants.PORTFOLIO_AVAILABLE: 0,
                            commons_constants.PORTFOLIO_TOTAL: 0
                        }
                    ending_portfolio[asset][commons_constants.PORTFOLIO_AVAILABLE] += \
                        value[commons_constants.PORTFOLIO_AVAILABLE]
                    ending_portfolio[asset][commons_constants.PORTFOLIO_TOTAL] += \
                        value[commons_constants.PORTFOLIO_TOTAL]
            current_value += latest_value[portfolio_history.HistoricalAssetValue.VALUES_KEY].get(reference_market, 0)
            min_ts = first_value[portfolio_history.HistoricalAssetValue.TIMESTAMP_KEY]
            initial_value_by_timestamp[min_ts] = initial_value_by_timestamp.get(min_ts, 0) + \
                first_value[portfolio_history.HistoricalAssetValue.VALUES_KEY].get(reference_market, 0)
            history = [
                history_val
                for history_val in hist_portfolio_values_manager.get_dict_historical_values()
                if history_val[portfolio_history.HistoricalAssetValue.TIMESTAMP_KEY] in self._to_update_auth_data_ids_buffer
            ]
            if not self.exchange_manager.is_trader_simulated:
                # skip portfolio history on simulated trading
                for history_val in history:
                    ts = history_val[portfolio_history.HistoricalAssetValue.TIMESTAMP_KEY]
                    value = history_val[portfolio_history.HistoricalAssetValue.VALUES_KEY]
                    if ts in historical_value_by_timestamp:
                        for key, val in value.items():
                            if key in historical_value_by_timestamp[ts]:
                                historical_value_by_timestamp[ts][key] += val
                            else:
                                historical_value_by_timestamp[ts][key] = value
                    else:
                        historical_value_by_timestamp[ts] = value

        if current_value:
            # only consider the initial value of the min timestamp
            initial_value = initial_value_by_timestamp[min(initial_value_by_timestamp)]
            await authenticator.update_portfolio(
                {reference_market: current_value},
                {reference_market: initial_value},
                ((100 * current_value / initial_value) - 100) if initial_value else 0,
                reference_market,
                ending_portfolio,
                historical_value_by_timestamp,
                price_by_asset,
                reset
            )
            self._to_update_auth_data_ids_buffer.clear()

    def get_db(self):
        return self._get_db()

    def _get_db(self):
        return commons_databases.RunDatabasesProvider.instance().get_historical_portfolio_value_db(
            self.exchange_manager.bot_id,
            storage_util.get_account_type_suffix_from_exchange_manager(self.exchange_manager),
            self.exchange_manager.exchange_name,
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.orders import order_state
from octobot_trading.personal_data.orders.order_state import (
    OrderState,
)
from octobot_trading.personal_data.orders import order_group
from octobot_trading.personal_data.orders.order_group import (
    OrderGroup,
)
from octobot_trading.personal_data.orders import groups
from octobot_trading.personal_data.orders.groups import (
    BalancedTakeProfitAndStopOrderGroup,
    OneCancelsTheOtherOrderGroup,
    get_group_class,
    get_or_create_order_group_from_storage_order_details,
)
from octobot_trading.personal_data.orders import order
from octobot_trading.personal_data.orders.order import (
    Order,
    parse_order_type,
)
from octobot_trading.personal_data.orders import types
from octobot_trading.personal_data.orders.types import (
    UnsupportedOrder,
    UnknownOrder,
    MarketOrder,
    SellMarketOrder,
    BuyMarketOrder,
    BuyLimitOrder,
    SellLimitOrder,
    LimitOrder,
    TakeProfitOrder,
    StopLossOrder,
    StopLossLimitOrder,
    TakeProfitLimitOrder,
    TrailingStopOrder,
    TrailingStopLimitOrder,
)
from octobot_trading.personal_data.orders import states
from octobot_trading.personal_data.orders.states import (
    CloseOrderState,
    CancelOrderState,
    OpenOrderState,
    create_order_state,
    FillOrderState,
    PendingCreationOrderState,
    PendingCreationChainedOrderState,
)
from octobot_trading.personal_data.orders import channel
from octobot_trading.personal_data.orders.channel import (
    OrdersProducer,
    OrdersChannel,
    OrdersUpdater,
    OrdersUpdaterSimulator,
)
from octobot_trading.personal_data.orders import orders_manager
from octobot_trading.personal_data.orders.orders_manager import (
    OrdersManager,
)
from octobot_trading.personal_data.orders import order_util
from octobot_trading.personal_data.orders.order_util import (
    is_valid,
    get_min_max_amounts,
    check_cost,
    get_valid_split_orders,
    get_split_orders_count_and_increment,
    get_futures_max_order_size,
    get_max_order_quantity_for_price,
    get_locked_funds,
    total_fees_from_order_dict,
    get_fees_for_currency,
    parse_raw_fees,
    parse_order_status,
    parse_is_cancelled,
    parse_is_open,
    get_up_to_date_price,
    get_pre_order_data,
    get_portfolio_amounts,
    get_pnl_transaction_source_from_order,
    is_stop_order,
    is_take_profit_order,
    get_trade_order_type,
    create_as_chained_order,
    is_associated_pending_order,
    apply_pending_order_from_created_order,
    ensure_orders_relevancy,
    get_order_quantity_currency,
    get_order_size_portfolio_percent,
    generate_order_id,
    wait_for_order_fill,
)
from octobot_trading.personal_data.orders import orders_storage_operations
from octobot_trading.personal_data.orders.orders_storage_operations import (
    apply_order_storage_details_if_any,
    create_orders_storage_related_elements,
    create_missing_self_managed_orders_from_storage_order_groups,
)
from octobot_trading.personal_data.orders import order_adapter
from octobot_trading.personal_data.orders.order_adapter import (
    adapt_price,
    adapt_quantity,
    trunc_with_n_decimal_digits,
    adapt_order_quantity_because_quantity,
    adapt_order_quantity_because_price,
    split_orders,
    check_and_adapt_order_details_if_necessary,
    add_dusts_to_quantity_if_necessary,
)
from octobot_trading.personal_data.orders.decimal_order_adapter import (
    get_minimal_order_amount,
    get_minimal_order_cost,
    decimal_adapt_price,
    decimal_adapt_quantity,
    decimal_trunc_with_n_decimal_digits,
    decimal_adapt_order_quantity_because_quantity,
    decimal_adapt_order_quantity_because_price,
    decimal_adapt_order_quantity_because_fees,
    decimal_split_orders,
    decimal_check_and_adapt_order_details_if_necessary,
    decimal_add_dusts_to_quantity_if_necessary,
)
from octobot_trading.personal_data.orders import order_factory
from octobot_trading.personal_data.orders.order_factory import (
    create_order_from_raw,
    create_order_instance_from_raw,
    create_order_from_type,
    create_order_instance,
    create_order_from_dict,
    create_order_from_order_storage_details,
)

__all__ = [
    "Order",
    "parse_order_type",
    "is_valid",
    "get_min_max_amounts",
    "check_cost",
    "get_valid_split_orders",
    "get_split_orders_count_and_increment",
    "get_futures_max_order_size",
    "get_max_order_quantity_for_price",
    "get_locked_funds",
    "total_fees_from_order_dict",
    "get_fees_for_currency",
    "parse_raw_fees",
    "parse_order_status",
    "parse_is_cancelled",
    "parse_is_open",
    "get_up_to_date_price",
    "get_pre_order_data",
    "get_portfolio_amounts",
    "get_pnl_transaction_source_from_order",
    "create_as_chained_order",
    "ensure_orders_relevancy",
    "get_order_quantity_currency",
    "get_order_size_portfolio_percent",
    "generate_order_id",
    "wait_for_order_fill",
    "apply_order_storage_details_if_any",
    "create_missing_self_managed_orders_from_storage_order_groups",
    "is_associated_pending_order",
    "apply_pending_order_from_created_order",
    "OrderState",
    "OrderGroup",
    "BalancedTakeProfitAndStopOrderGroup",
    "OneCancelsTheOtherOrderGroup",
    "get_group_class",
    "get_or_create_order_group_from_storage_order_details",
    "OrdersUpdater",
    "adapt_price",
    "get_minimal_order_amount",
    "get_minimal_order_cost",
    "decimal_adapt_price",
    "adapt_quantity",
    "decimal_adapt_quantity",
    "trunc_with_n_decimal_digits",
    "decimal_trunc_with_n_decimal_digits",
    "adapt_order_quantity_because_quantity",
    "decimal_adapt_order_quantity_because_quantity",
    "adapt_order_quantity_because_price",
    "decimal_adapt_order_quantity_because_price",
    "decimal_adapt_order_quantity_because_fees",
    "split_orders",
    "decimal_split_orders",
    "check_and_adapt_order_details_if_necessary",
    "decimal_check_and_adapt_order_details_if_necessary",
    "add_dusts_to_quantity_if_necessary",
    "decimal_add_dusts_to_quantity_if_necessary",
    "create_order_from_raw",
    "create_order_instance_from_raw",
    "create_order_from_type",
    "create_order_instance",
    "create_order_from_dict",
    "create_order_from_order_storage_details",
    "OrdersProducer",
    "OrdersChannel",
    "OrdersManager",
    "OrdersUpdaterSimulator",
    "CloseOrderState",
    "CancelOrderState",
    "OpenOrderState",
    "create_order_state",
    "FillOrderState",
    "PendingCreationOrderState",
    "PendingCreationChainedOrderState",
    "UnsupportedOrder",
    "UnknownOrder",
    "MarketOrder",
    "SellMarketOrder",
    "BuyMarketOrder",
    "BuyLimitOrder",
    "SellLimitOrder",
    "LimitOrder",
    "TakeProfitOrder",
    "StopLossOrder",
    "StopLossLimitOrder",
    "TakeProfitLimitOrder",
    "TrailingStopOrder",
    "TrailingStopLimitOrder",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import async_channel.constants as constants

import octobot_trading.exchange_channel as exchanges_channel


class OrderBookProducer(exchanges_channel.ExchangeChannelProducer):
    async def push(self, symbol, asks, bids, update_order_book=True):
        await self.perform(symbol, asks, bids, update_order_book)

    async def perform(self, symbol, asks, bids, update_order_book):
        try:
            if self.channel.get_filtered_consumers(symbol=constants.CHANNEL_WILDCARD) or \
                    self.channel.get_filtered_consumers(symbol=symbol):
                if update_order_book:
                    self.channel.exchange_manager.get_symbol_data(symbol).handle_order_book_update(asks, bids)
                await self.send(cryptocurrency=self.channel.exchange_manager.exchange.
                                get_pair_cryptocurrency(symbol),
                                symbol=symbol,
                                asks=asks,
                                bids=bids)
        except asyncio.CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.exception(e, True, f"Exception when triggering update: {e}")

    async def send(self, cryptocurrency, symbol, asks, bids):
        for consumer in self.channel.get_filtered_consumers(symbol=symbol):
            await consumer.queue.put({
                "exchange": self.channel.exchange_manager.exchange_name,
                "exchange_id": self.channel.exchange_manager.id,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "asks": asks,
                "bids": bids
            })


class OrderBookChannel(exchanges_channel.ExchangeChannel):
    PRODUCER_CLASS = OrderBookProducer
    CONSUMER_CLASS = exchanges_channel.ExchangeChannelConsumer


class OrderBookTickerProducer(exchanges_channel.ExchangeChannelProducer):
    async def push(self, symbol, ask_quantity, ask_price, bid_quantity, bid_price):
        await self.perform(symbol, ask_quantity, ask_price, bid_quantity, bid_price)

    async def perform(self, symbol, ask_quantity, ask_price, bid_quantity, bid_price):
        try:
            if self.channel.get_filtered_consumers(symbol=constants.CHANNEL_WILDCARD) or \
                    self.channel.get_filtered_consumers(symbol=symbol):
                self.channel.exchange_manager.get_symbol_data(symbol).handle_order_book_ticker_update(ask_quantity,
                                                                                                      ask_price,
                                                                                                      bid_quantity,
                                                                                                      bid_price)
                await self.send(cryptocurrency=self.channel.exchange_manager.exchange.
                                get_pair_cryptocurrency(symbol),
                                symbol=symbol,
                                ask_quantity=ask_quantity, ask_price=ask_price,
                                bid_quantity=bid_quantity, bid_price=bid_price)
        except asyncio.CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.exception(e, True, f"Exception when triggering update: {e}")

    async def send(self, cryptocurrency, symbol, ask_quantity, ask_price, bid_quantity, bid_price):
        for consumer in self.channel.get_filtered_consumers(symbol=symbol):
            await consumer.queue.put({
                "exchange": self.channel.exchange_manager.exchange_name,
                "exchange_id": self.channel.exchange_manager.id,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "ask_quantity": ask_quantity,
                "ask_price": ask_price,
                "bid_quantity": bid_quantity,
                "bid_price": bid_price
            })


class OrderBookTickerChannel(exchanges_channel.ExchangeChannel):
    PRODUCER_CLASS = OrderBookTickerProducer
    CONSUMER_CLASS = exchanges_channel.ExchangeChannelConsumer



#  Drakkar-Software OctoBot-Private-Tentacles
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal

import async_channel.enums as channel_enums
import octobot_commons.channels_name as channels_name

import octobot_trading.enums as trading_enums
import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.personal_data as trading_personal_data
import octobot_trading.util.test_tools.exchanges_test_tools as exchanges_test_tools


class SpotRestExchangeTests:
    SUBSCRIBED_CHANNELS = [channels_name.OctoBotTradingChannelsName.BALANCE_CHANNEL.value,
                           channels_name.OctoBotTradingChannelsName.TRADES_CHANNEL.value,
                           channels_name.OctoBotTradingChannelsName.ORDERS_CHANNEL.value]

    def __init__(self, exchange_name, config):
        self.config = config
        self.exchange_name = exchange_name

        self.exchange_manager_instance = None

        self.channel_callbacks_triggered = {}
        self.expected_crypto_in_balance = []

    async def initialize(self):
        self.exchange_manager_instance = await exchanges_test_tools.create_test_exchange_manager(
            config=self.config, exchange_name=self.exchange_name,
            rest_only=True, is_spot_only=True, is_sandboxed=True, ignore_exchange_config=False)
        await self.exchange_manager_instance.trader.initialize()

    async def run(self, symbol):
        await self.subscribe_personal_channels()
        await self.exchange_manager_instance.trader.cancel_all_open_orders()
        await self.test_orders(symbol)

    async def stop(self):
        await exchanges_test_tools.stop_test_exchange_manager(self.exchange_manager_instance)

    async def test_orders(self, symbol):
        mark_price = await self.exchange_manager_instance.exchange_symbols_data.get_exchange_symbol_data(symbol) \
            .prices_manager.get_mark_price(timeout=60)
        await self._create_order(symbol,
                                 trading_enums.TraderOrderType.SELL_LIMIT,
                                 mark_price * decimal.Decimal("1.1"),
                                 decimal.Decimal("1.001"),
                                 mark_price * decimal.Decimal("1.1"))
        await asyncio.sleep(7)
        assert (await self._get_exchange_orders_count(symbol)) == 1
        open_orders = self._get_orders_manager_open_orders(symbol)
        assert len(open_orders) == 1
        await self.exchange_manager_instance.trader.cancel_order_with_id(open_orders[-1].order_id)
        await asyncio.sleep(7)
        assert (await self._get_exchange_orders_count(symbol)) == 0
        open_orders = self._get_orders_manager_open_orders(symbol)
        assert len(open_orders) == 0

    async def _get_exchange_orders_count(self, symbol):
        return len(await self.exchange_manager_instance.exchange.get_open_orders(symbol=symbol))

    def _get_orders_manager_open_orders(self, symbol):
        return self.exchange_manager_instance.exchange_personal_data.orders_manager.get_open_orders(symbol)

    async def test_all_callback_triggered(self):
        for channel in self.SUBSCRIBED_CHANNELS:
            assert self.channel_callbacks_triggered.get(channel, False)

    async def _create_order(self, symbol, order_type, current_price, quantity, limit_price):
        current_order = trading_personal_data.create_order_instance(trader=self.exchange_manager_instance.trader,
                                                                    order_type=order_type,
                                                                    symbol=symbol,
                                                                    current_price=current_price,
                                                                    quantity=quantity,
                                                                    price=limit_price)
        await self.exchange_manager_instance.trader.create_order(current_order)

    async def subscribe_personal_channels(self):
        await exchanges_channel.get_chan(channels_name.OctoBotTradingChannelsName.BALANCE_CHANNEL.value,
                                         self.exchange_manager_instance.id).new_consumer(
            self._balance_callback, priority_level=channel_enums.ChannelConsumerPriorityLevels.MEDIUM.value
        )
        await exchanges_channel.get_chan(channels_name.OctoBotTradingChannelsName.TRADES_CHANNEL.value,
                                         self.exchange_manager_instance.id).new_consumer(
            self._trades_callback, priority_level=channel_enums.ChannelConsumerPriorityLevels.MEDIUM.value
        )
        await exchanges_channel.get_chan(channels_name.OctoBotTradingChannelsName.ORDERS_CHANNEL.value,
                                         self.exchange_manager_instance.id).new_consumer(
            self._orders_callback, priority_level=channel_enums.ChannelConsumerPriorityLevels.MEDIUM.value
        )

    def _check_balance_crypto(self, balance):
        for crypto in self.expected_crypto_in_balance:
            assert crypto in balance

    async def _balance_callback(self, exchange: str, exchange_id: str, balance):
        self.channel_callbacks_triggered[channels_name.OctoBotTradingChannelsName.BALANCE_CHANNEL.value] = True
        self._check_balance_crypto(balance)

    async def _trades_callback(
            self,
            exchange: str,
            exchange_id: str,
            cryptocurrency: str,
            symbol: str,
            trade: dict,
            old_trade: bool,
    ):
        self.channel_callbacks_triggered[channels_name.OctoBotTradingChannelsName.TRADES_CHANNEL.value] = True

    async def _orders_callback(
            self,
            exchange: str,
            exchange_id: str,
            cryptocurrency: str,
            symbol: str,
            order: dict,
            update_type: str,
            is_from_bot: bool,
    ):
        self.channel_callbacks_triggered[channels_name.OctoBotTradingChannelsName.ORDERS_CHANNEL.value] = True



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock
import decimal

import octobot_trading.personal_data as trading_personal_data
import octobot_trading.enums as trading_enums
import octobot_trading.modes.script_keywords as script_keywords
import octobot_trading.modes.script_keywords.basic_keywords.account_balance as account_balance
import octobot_trading.constants as trading_constants
import octobot_commons.constants as commons_constants

from tests import event_loop
from tests.modes.script_keywords import null_context, mock_context
from tests.exchanges import \
    backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_total_account_balance(mock_context):
    mock_context.exchange_manager.exchange_personal_data. \
        portfolio_manager.portfolio_value_holder.portfolio_current_value = decimal.Decimal(999)
    mock_context.exchange_manager.exchange_personal_data.portfolio_manager.reference_market = "USDT"
    with mock.patch.object(trading_personal_data, "get_up_to_date_price", mock.AsyncMock(return_value=100)) \
            as get_up_to_date_price:
        assert await script_keywords.total_account_balance(mock_context) == decimal.Decimal("9.99")
        get_up_to_date_price.assert_called_once_with(mock_context.exchange_manager, symbol=mock_context.symbol,
                                                     timeout=trading_constants.ORDER_DATA_FETCHING_TIMEOUT)


async def test_available_account_balance(mock_context):
    current_symbol_holding = decimal.Decimal("100")
    market_quantity = decimal.Decimal("50")

    async def _get_pre_order_data(
        exchange_manager, symbol: str, timeout: int = None,
        portfolio_type=commons_constants.PORTFOLIO_AVAILABLE,
        target_price=None
    ):
        if portfolio_type == commons_constants.PORTFOLIO_AVAILABLE:
            return current_symbol_holding, "current_market_holding", market_quantity, "current_price", "symbol_market"
        return current_symbol_holding*2, "current_market_holding", market_quantity*2, "current_price", "symbol_market"

    with mock.patch.object(trading_personal_data, "get_pre_order_data", mock.AsyncMock(side_effect=_get_pre_order_data)) \
            as get_pre_order_data_mock, \
            mock.patch.object(account_balance, "_get_locked_amount_in_stop_orders",
                              mock.Mock(return_value=trading_constants.ONE)) as _get_locked_amount_in_stop_orders_mock:
        assert market_quantity == await script_keywords.available_account_balance(
            mock_context, side=trading_enums.TradeOrderSide.BUY.value, use_total_holding=False, is_stop_order=False,
            target_price=trading_constants.ZERO
        )
        get_pre_order_data_mock.assert_called_once_with(mock_context.exchange_manager, symbol=mock_context.symbol,
                                                        timeout=trading_constants.ORDER_DATA_FETCHING_TIMEOUT,
                                                        portfolio_type=commons_constants.PORTFOLIO_AVAILABLE,
                                                        target_price=trading_constants.ZERO)
        get_pre_order_data_mock.reset_mock()
        _get_locked_amount_in_stop_orders_mock.assert_not_called()
        assert current_symbol_holding * 2 == await script_keywords.available_account_balance(
            mock_context, side=trading_enums.TradeOrderSide.SELL.value, use_total_holding=True)
        get_pre_order_data_mock.assert_called_once_with(mock_context.exchange_manager, symbol=mock_context.symbol,
                                                        timeout=trading_constants.ORDER_DATA_FETCHING_TIMEOUT,
                                                        portfolio_type=commons_constants.PORTFOLIO_TOTAL,
                                                        target_price=None)
        get_pre_order_data_mock.reset_mock()
        _get_locked_amount_in_stop_orders_mock.assert_not_called()
        assert current_symbol_holding * 2 - trading_constants.ONE == await script_keywords.available_account_balance(
            mock_context, side=trading_enums.TradeOrderSide.SELL.value, use_total_holding=True, is_stop_order=True)
        get_pre_order_data_mock.assert_called_once_with(mock_context.exchange_manager, symbol=mock_context.symbol,
                                                        timeout=trading_constants.ORDER_DATA_FETCHING_TIMEOUT,
                                                        portfolio_type=commons_constants.PORTFOLIO_TOTAL,
                                                        target_price=None)
        _get_locked_amount_in_stop_orders_mock.assert_called_once_with(mock_context,
                                                                       trading_enums.TradeOrderSide.SELL.value)
        get_pre_order_data_mock.reset_mock()
        _get_locked_amount_in_stop_orders_mock.reset_mock()
        assert market_quantity * 2 - trading_constants.ONE == await script_keywords.available_account_balance(
            mock_context, side=trading_enums.TradeOrderSide.BUY.value, use_total_holding=True, is_stop_order=True)
        get_pre_order_data_mock.assert_called_once_with(mock_context.exchange_manager, symbol=mock_context.symbol,
                                                        timeout=trading_constants.ORDER_DATA_FETCHING_TIMEOUT,
                                                        portfolio_type=commons_constants.PORTFOLIO_TOTAL,
                                                        target_price=None)
        _get_locked_amount_in_stop_orders_mock.assert_called_once_with(mock_context,
                                                                       trading_enums.TradeOrderSide.BUY.value)
        get_pre_order_data_mock.reset_mock()
        _get_locked_amount_in_stop_orders_mock.reset_mock()

        buy_1 = trading_personal_data.BuyLimitOrder(mock_context.trader)
        buy_1.origin_quantity = decimal.Decimal("10")
        buy_1.origin_price = decimal.Decimal("100")
        buy_2 = trading_personal_data.BuyLimitOrder(mock_context.trader)
        buy_2.origin_quantity = decimal.Decimal("3")
        buy_2.origin_price = decimal.Decimal("40")
        sell_1 = trading_personal_data.SellLimitOrder(mock_context.trader)
        sell_1.origin_quantity = decimal.Decimal("2.21")
        sell_1.origin_price = decimal.Decimal("200")
        assert market_quantity + buy_1.origin_quantity + buy_2.origin_quantity == \
           await script_keywords.available_account_balance(
                mock_context, side=trading_enums.TradeOrderSide.BUY.value, use_total_holding=False, is_stop_order=False,
                orders_to_be_ignored=[buy_1, buy_2, sell_1]
            )
        assert get_pre_order_data_mock.call_count == 2
        _get_locked_amount_in_stop_orders_mock.assert_not_called()
        get_pre_order_data_mock.reset_mock()
        _get_locked_amount_in_stop_orders_mock.reset_mock()


async def test_adapt_amount_to_holdings(null_context):
    with mock.patch.object(account_balance, "available_account_balance",
                           mock.AsyncMock(return_value=decimal.Decimal(1))) as available_account_balance_mock:
        assert await script_keywords.adapt_amount_to_holdings(null_context,
                                                              decimal.Decimal(0),
                                                              trading_enums.TradeOrderSide.SELL,
                                                              False,
                                                              "reduce_only",
                                                              "is_stop_order",
                                                              target_price=trading_constants.ONE_HUNDRED
                                                              ) == decimal.Decimal(0)
        available_account_balance_mock.assert_called_once_with(null_context, trading_enums.TradeOrderSide.SELL,
                                                               use_total_holding=False, is_stop_order="is_stop_order",
                                                               reduce_only="reduce_only",
                                                               target_price=trading_constants.ONE_HUNDRED,
                                                               orders_to_be_ignored=None)
        available_account_balance_mock.reset_mock()
        assert await script_keywords.adapt_amount_to_holdings(null_context,
                                                              decimal.Decimal(1),
                                                              trading_enums.TradeOrderSide.SELL,
                                                              True,
                                                              "reduce_only",
                                                              "is_stop_order",
                                                              orders_to_be_ignored=["plop"]) == decimal.Decimal(1)
        available_account_balance_mock.assert_called_once_with(null_context, trading_enums.TradeOrderSide.SELL,
                                                               use_total_holding=True, is_stop_order="is_stop_order",
                                                               reduce_only="reduce_only",
                                                               target_price=None,
                                                               orders_to_be_ignored=["plop"])
        assert await script_keywords.adapt_amount_to_holdings(null_context,
                                                              decimal.Decimal(2),
                                                              trading_enums.TradeOrderSide.SELL,
                                                              False,
                                                              "reduce_only",
                                                              "is_stop_order") == decimal.Decimal(1)


async def test_get_locked_amount_in_stop_orders(mock_context):
    mock_context.exchange_manager = mock.Mock()
    mock_context.exchange_manager.exchange_personal_data = mock.Mock()
    mock_context.exchange_manager.exchange_personal_data.orders_manager = mock.Mock()

    mock_context.exchange_manager.exchange_personal_data.orders_manager.get_open_orders = mock.Mock(return_value=[])
    assert account_balance._get_locked_amount_in_stop_orders(mock_context, "buy") == trading_constants.ZERO

    order = trading_personal_data.StopLossOrder(mock_context.trader)
    order.side = trading_enums.TradeOrderSide.SELL
    order.origin_quantity = decimal.Decimal("10")

    mock_context.exchange_manager.exchange_personal_data.orders_manager.get_open_orders = \
        mock.Mock(return_value=[order])
    assert account_balance._get_locked_amount_in_stop_orders(mock_context, "buy") == trading_constants.ZERO

    mock_context.exchange_manager.exchange_personal_data.orders_manager.get_open_orders = \
        mock.Mock(return_value=[order])
    assert account_balance._get_locked_amount_in_stop_orders(mock_context, "sell") == decimal.Decimal("10")

    order_2 = trading_personal_data.StopLossLimitOrder(mock_context.trader)
    order_2.side = trading_enums.TradeOrderSide.SELL
    order_2.origin_quantity = decimal.Decimal("20")

    order_3 = trading_personal_data.StopLossLimitOrder(mock_context.trader)
    order_3.side = trading_enums.TradeOrderSide.BUY
    order_3.origin_quantity = decimal.Decimal("50")

    order_4 = trading_personal_data.BuyLimitOrder(mock_context.trader)
    order_4.side = trading_enums.TradeOrderSide.SELL
    order_4.origin_quantity = decimal.Decimal("100")

    mock_context.exchange_manager.exchange_personal_data.orders_manager.get_open_orders = \
        mock.Mock(return_value=[order, order_2, order_3, order_4])
    assert account_balance._get_locked_amount_in_stop_orders(mock_context, "sell") == decimal.Decimal("30")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import mock
import pytest

import octobot_trading.personal_data as personal_data
import octobot_trading.personal_data.orders.groups.balanced_take_profit_and_stop_order_group as \
    balanced_take_profit_and_stop_order_group
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.order_util as order_util
from tests.exchanges import backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests.personal_data.orders.groups import order_mock


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def btps_group(backtesting_exchange_manager):
    orders_manager = mock.Mock()
    orders_manager.get_order_from_group = mock.Mock()
    return personal_data.BalancedTakeProfitAndStopOrderGroup("name",  orders_manager)


@pytest.fixture
def side_balance():
    return balanced_take_profit_and_stop_order_group._SideBalance()


async def test_on_fill(btps_group):
    order = order_mock()
    # mock btps_group.orders_manager instead of btps_group._balance_orders to allow cythonized class tests
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
            as get_order_from_group_mock:
        await btps_group.on_fill(order)
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()

        await btps_group.on_fill(order, ["hi", "ho"])
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()


async def test_on_cancel(btps_group):
    order = order_mock()
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
            as get_order_from_group_mock:
        await btps_group.on_cancel(order)
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()

        await btps_group.on_cancel(order, ["hi", "ho"])
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()


async def test_enable(btps_group):
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
            as get_order_from_group_mock:
        await btps_group.enable(False)
        get_order_from_group_mock.assert_not_called()
        await btps_group.enable(True)
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        await btps_group.enable(True)
        get_order_from_group_mock.assert_called_once()


async def test_can_create_order(btps_group):
    order_1 = order_mock(origin_quantity=decimal.Decimal(1), order_type=enums.TraderOrderType.SELL_LIMIT,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    order_2 = order_mock(origin_quantity=decimal.Decimal(5), order_type=enums.TraderOrderType.BUY_MARKET,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    order_3 = order_mock(origin_quantity=decimal.Decimal(10), order_type=enums.TraderOrderType.TRAILING_STOP,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                           mock.Mock(return_value=[])) as get_order_from_group_mock:
        # 0 size order
        assert btps_group.can_create_order(enums.TraderOrderType.STOP_LOSS, decimal.Decimal(0)) is True
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        # no order, no imbalance
        assert btps_group.can_create_order(enums.TraderOrderType.STOP_LOSS, decimal.Decimal(1)) is False
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                           mock.Mock(return_value=[order_1, order_2, order_3])) as get_order_from_group_mock:
        # enough imbalance
        assert btps_group.can_create_order(enums.TraderOrderType.SELL_LIMIT, decimal.Decimal(1)) is True
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        assert btps_group.can_create_order(enums.TraderOrderType.SELL_LIMIT, decimal.Decimal(4)) is True
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()

        # not enough
        assert btps_group.can_create_order(enums.TraderOrderType.SELL_LIMIT, decimal.Decimal("4.000001")) is False
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        assert btps_group.can_create_order(enums.TraderOrderType.STOP_LOSS, decimal.Decimal("0.00001")) is False
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()


async def test_get_max_order_quantity(btps_group):
    order_1 = order_mock(origin_quantity=decimal.Decimal(1), order_type=enums.TraderOrderType.SELL_LIMIT,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    order_2 = order_mock(origin_quantity=decimal.Decimal(5), order_type=enums.TraderOrderType.BUY_MARKET,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    order_3 = order_mock(origin_quantity=decimal.Decimal(10), order_type=enums.TraderOrderType.TRAILING_STOP,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                           mock.Mock(return_value=[])) as get_order_from_group_mock:
        # no order, no imbalance
        assert btps_group.get_max_order_quantity(enums.TraderOrderType.STOP_LOSS) == constants.ZERO
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                           mock.Mock(return_value=[order_1, order_2, order_3])) as get_order_from_group_mock:
        # imbalance
        assert btps_group.get_max_order_quantity(enums.TraderOrderType.SELL_LIMIT) == decimal.Decimal(4)
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        assert btps_group.get_max_order_quantity(enums.TraderOrderType.STOP_LOSS) == constants.ZERO
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()


async def test_balance_orders(btps_group):
    order_1 = order_mock(origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(10),
                         order_type=enums.TraderOrderType.SELL_LIMIT, origin_quantity=decimal.Decimal("1.5"))
    order_2 = order_mock(origin_price=decimal.Decimal(5), created_last_price=decimal.Decimal(3),
                         order_type=enums.TraderOrderType.SELL_LIMIT, origin_quantity=decimal.Decimal("11"))
    order_3 = order_mock(origin_price=decimal.Decimal(10), created_last_price=decimal.Decimal(11),
                         order_type=enums.TraderOrderType.STOP_LOSS, origin_quantity=decimal.Decimal("10"))
    order_4 = order_mock(origin_price=decimal.Decimal(42), created_last_price=decimal.Decimal(2),
                         order_type=enums.TraderOrderType.STOP_LOSS, origin_quantity=decimal.Decimal("20"))
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                           mock.Mock(return_value=[order_1, order_2, order_3, order_4])) as get_order_from_group_mock:
        base_balancing_order = ["existing_order"]
        btps_group.balancing_orders = base_balancing_order
        # 1. not enabled
        btps_group.enabled = False
        await btps_group._balance_orders(order_1, ["ignored_orders"])
        get_order_from_group_mock.assert_not_called()
        assert btps_group.balancing_orders == ["existing_order"]
        assert btps_group.balancing_orders is base_balancing_order

        # 2. enabled
        btps_group.enabled = True
        await btps_group._balance_orders(order_4, ["ignored_orders"])
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        order_1.trader.edit_order.assert_not_called()
        order_1.trader.cancel_order.assert_called_once_with(
            order_1, ignored_order=order_4, wait_for_cancelling=True,
            cancelling_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
        )
        order_2.trader.edit_order.assert_called_once_with(
            order_2,
            edited_quantity=decimal.Decimal(10),
            edited_price=None,
            edited_stop_price=None,
            edited_current_price=None,
            params=None
        )
        order_2.trader.cancel_order.assert_not_called()
        order_3.trader.edit_order.assert_not_called()
        order_3.trader.cancel_order.assert_not_called()
        order_4.trader.edit_order.assert_not_called()
        order_4.trader.cancel_order.assert_not_called()
        assert btps_group.balancing_orders == ["existing_order"]
        assert btps_group.balancing_orders is not base_balancing_order  # changed list during orders registration


async def test_get_balance(btps_group):
    order_1 = order_mock(order_type="order_type", origin_price=decimal.Decimal(1),
                         created_last_price=decimal.Decimal(2))
    order_2 = order_mock(order_type="order_type", origin_price=decimal.Decimal(1),
                         created_last_price=decimal.Decimal(2))
    with mock.patch.object(order_util, "is_stop_order", mock.Mock(return_value=False)) as is_stop_order_mock:
        with mock.patch.object(btps_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
             as get_order_from_group_mock:
            balance = btps_group._get_balance(order_1, None)
            assert balance[btps_group.TAKE_PROFIT].orders == []
            assert balance[btps_group.STOP].orders == []
            get_order_from_group_mock.assert_called_once()
            is_stop_order_mock.assert_not_called()

        with mock.patch.object(btps_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[order_1, order_2])) \
             as get_order_from_group_mock:
            balance = btps_group._get_balance(order_1, None)
            assert balance[btps_group.TAKE_PROFIT].orders == [order_2]
            assert balance[btps_group.STOP].orders == []
            get_order_from_group_mock.assert_called_once()
            is_stop_order_mock.assert_called_once_with("order_type")
            is_stop_order_mock.reset_mock()

        with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                               mock.Mock(return_value=[order_1, order_2])) as get_order_from_group_mock:
            balance = btps_group._get_balance(order_1, [order_2])
            assert balance[btps_group.TAKE_PROFIT].orders == []
            assert balance[btps_group.STOP].orders == []
            get_order_from_group_mock.assert_called_once()
            is_stop_order_mock.assert_not_called()

    with mock.patch.object(order_util, "is_stop_order", mock.Mock(return_value=True)) as is_stop_order_mock:
        with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                               mock.Mock(return_value=[order_1, order_2])) as get_order_from_group_mock:
            balance = btps_group._get_balance(order_1, None)
            assert balance[btps_group.TAKE_PROFIT].orders == []
            assert balance[btps_group.STOP].orders == [order_2]
            get_order_from_group_mock.assert_called_once()
            is_stop_order_mock.assert_called_once_with("order_type")
            is_stop_order_mock.reset_mock()


async def test_SideBalance_add_order(side_balance):
    assert side_balance.orders == []
    order_1 = order_mock(origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    order_2 = order_mock(origin_price=decimal.Decimal(15), created_last_price=decimal.Decimal(3))
    order_3 = order_mock(origin_price=decimal.Decimal(50), created_last_price=decimal.Decimal(55))
    side_balance.add_order(order_1)
    assert side_balance.orders == [order_1]
    side_balance.add_order(order_2)
    assert side_balance.orders == [order_2, order_1]
    side_balance.add_order(order_3)
    assert side_balance.orders == [order_2, order_3, order_1]


async def test_SideBalance_get_actions_to_balance_special_inputs(side_balance):
    with mock.patch.object(side_balance, "get_balance", mock.Mock(return_value=decimal.Decimal("-1"))):
        # target balance can't be negative (order quantity is always positive). value treated as abs()
        assert side_balance.get_actions_to_balance(constants.ONE) == {
            personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [],
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
        }
    with mock.patch.object(side_balance, "get_balance", mock.Mock(return_value=constants.ONE)):
        # target balance == balance: nothing to do
        assert side_balance.get_actions_to_balance(constants.ONE) == {
            personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [],
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
        }
    with mock.patch.object(side_balance, "get_balance", mock.Mock(return_value=constants.ONE)):
        # target balance > balance: nothing to do (can't create order to increase)
        assert side_balance.get_actions_to_balance(decimal.Decimal("0.6")) == {
            personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [],
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
        }
    with mock.patch.object(side_balance, "get_balance", mock.Mock(return_value=constants.ONE)):
        # no order to create actions
        assert side_balance.get_actions_to_balance(decimal.Decimal("10")) == {
            personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [],
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
        }


async def test_SideBalance_get_actions_to_balance_normal_inputs(side_balance):
    order_0_1 = order_mock(origin_quantity=decimal.Decimal("2.899999"))
    order_0_2 = order_mock(origin_quantity=decimal.Decimal("0.111111"))
    order_0_3 = order_mock(origin_quantity=decimal.Decimal("3.15545145441"))
    side_balance.orders = [order_0_1, order_0_2, order_0_3]
    assert side_balance.get_actions_to_balance(decimal.Decimal("3.15545145441")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_0_1, order_0_2],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
    }
    order_1 = order_mock(origin_quantity=decimal.Decimal("1"))
    order_2 = order_mock(origin_quantity=decimal.Decimal("6.42"))
    side_balance.orders = [order_2]
    assert side_balance.get_actions_to_balance(decimal.Decimal("3")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_2,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: decimal.Decimal("3"),
        }],
    }
    side_balance.orders = [order_1, order_2]
    assert side_balance.get_actions_to_balance(decimal.Decimal("3")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_1],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_2,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: decimal.Decimal("3"),
        }],
    }
    order_2 = order_mock(origin_quantity=decimal.Decimal("6.42"))
    order_3 = order_mock(origin_quantity=decimal.Decimal("11.888998327557457"))
    side_balance.orders = [order_1, order_2, order_3]
    assert side_balance.get_actions_to_balance(decimal.Decimal("3")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_1, order_2],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_3,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: decimal.Decimal("3"),
        }],
    }
    order_4 = order_mock(origin_quantity=decimal.Decimal("0.1"))
    side_balance.orders = [order_1, order_2, order_3, order_4]
    # keep order_4 untouched as it is the closest from price and can be included in target balance
    assert side_balance.get_actions_to_balance(decimal.Decimal("3")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_1, order_2],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_3,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: decimal.Decimal("2.9"),
        }],
    }
    side_balance.orders = [order_1, order_4, order_2, order_3]
    assert side_balance.get_actions_to_balance(decimal.Decimal("3")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_1, order_4, order_2],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_3,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: decimal.Decimal("3"),
        }],
    }

    side_balance.orders = [order_1, order_4, order_2, order_3]
    assert side_balance.get_actions_to_balance(constants.ZERO) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_1, order_4, order_2, order_3],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
    }


async def test_SideBalance_get_balance(side_balance):
    assert side_balance.get_balance() is constants.ZERO
    order_1 = order_mock(origin_quantity=decimal.Decimal(1))
    order_2 = order_mock(origin_quantity=decimal.Decimal(10))
    order_mock(origin_quantity=decimal.Decimal(50))
    side_balance.orders = [order_1, order_2]
    assert side_balance.get_balance() == decimal.Decimal(11)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy


class HistoricalAssetValue:
    """
    HistoricalAssetValue stores the portfolio value at a given time in multiple currencies
    """
    TIMESTAMP_KEY = "t"
    VALUES_KEY = "v"
    SIGNIFICANT_VALUE_CHANGE_THRESHOLD = 0.1    # 10%

    def __init__(self, timestamp, value_by_currency):
        self._timestamp = timestamp
        self._value_by_currency = copy.copy(value_by_currency)

    def __contains__(self, item):
        return item in self._value_by_currency

    def __repr__(self):
        return f"{self.__class__.__name__} " \
               f"[timestamp: {HistoricalAssetValue}, _value_by_currency{self._value_by_currency}]"

    def get(self, currency):
        return self._value_by_currency[currency]

    def set(self, currency, value):
        self._value_by_currency[currency] = value

    def update(self, value_by_currency):
        # update exiting values and add new ones
        if self._value_by_currency == value_by_currency:
            return False
        self._value_by_currency.update(value_by_currency)
        return True

    def get_currencies(self):
        return self._value_by_currency.keys()

    def get_timestamp(self):
        return self._timestamp

    def is_significant_change(self, currency, value):
        stored_value = self.get(currency)
        if not stored_value:
            return bool(value)
        return abs(stored_value - value) / stored_value >= self.SIGNIFICANT_VALUE_CHANGE_THRESHOLD

    def to_dict(self):
        return {
            self.TIMESTAMP_KEY: self._timestamp,
            self.VALUES_KEY: {currency: float(value) for currency, value in self._value_by_currency.items()}
        }



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal

import octobot_commons.logging as logging
from octobot_trading.enums import ExchangeConstantsOrderColumns as ECOC


class PriceEventsManager:
    """
    Manage price events for a specific price and timestamp
    Mainly used for updating Order status
    """

    """
    The price event index from a price event tuple
    """
    PRICE_EVENT_INDEX = 2
    PRICE_KEY = "price"
    TIME_KEY = "time"
    MAX_LAST_RECENT_PRICES = 50

    def __init__(self):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.events = []
        self._last_recent_prices = []

    def stop(self):
        self.reset()

    def reset(self):
        """
        Reset price events
        """
        self.clear_recent_prices()
        self.events.clear()

    def get_min_and_max_prices(self) -> (float, float):
        if len(self._last_recent_prices) < 2:
            raise IndexError("Not enough data")
        prices = sorted([element[self.PRICE_KEY] for element in self._last_recent_prices])
        return prices[0], prices[-1]

    def handle_recent_trades(self, recent_trades):
        """
        Handle new recent trades prices
        :param recent_trades: prices to check
        """
        # reset recent prices on new recent trades
        self.clear_recent_prices()
        for recent_trade in recent_trades:
            price = decimal.Decimal(str(recent_trade[ECOC.PRICE.value]))
            timestamp = recent_trade[ECOC.TIMESTAMP.value]
            try:
                self._add_recent_price(price, timestamp)
                for event_to_set in self._check_events(price, timestamp):
                    self._remove_and_set_event(event_to_set)
            except KeyError:
                self.logger.error("Error when checking price events with recent trades data")

    def handle_price(self, price, timestamp):
        """
        Handle new simple price with timestamp
        :param price: the price to check
        :param timestamp: the timestamp to check
        """
        self._add_recent_price(price, timestamp)
        for event_to_set in self._check_events(price, timestamp):
            self._remove_and_set_event(event_to_set)

    def clear_recent_prices(self):
        self._last_recent_prices = []

    def _add_recent_price(self, price, timestamp):
        self._last_recent_prices.append({
            self.PRICE_KEY: price,
            self.TIME_KEY: timestamp
        })
        self._ensure_last_prices_length()

    def _ensure_last_prices_length(self):
        if len(self._last_recent_prices) > self.MAX_LAST_RECENT_PRICES:
            # only take the most recent prices
            self._last_recent_prices = self._last_recent_prices[self.MAX_LAST_RECENT_PRICES // 2:]

    def new_event(self, price, timestamp, trigger_above, allow_instant_fill=True):
        """
        Create a new event at price and timestamp.
        This event can already be set should it be instantly triggered.
        Otherwise it will be set once the required price conditions are met
        :param price: the trigger price
        :param timestamp: the timestamp to wait for
        :param trigger_above: True if waiting for an upper price
        :param allow_instant_fill: True if recent prices should be checked to fill this event
        :return: the price event
        """
        price_event_tuple = _new_price_event(price, timestamp, trigger_above)
        if allow_instant_fill and self._is_triggered_by_last_recent_prices(price, timestamp, trigger_above):
            # don't add to self.events an event that is already set
            price_event_tuple[PriceEventsManager.PRICE_EVENT_INDEX].set()
        else:
            # this event will be set when conditions are met
            self.events.append(price_event_tuple)
        return price_event_tuple[PriceEventsManager.PRICE_EVENT_INDEX]

    def _is_triggered_by_last_recent_prices(self, price, timestamp, trigger_above):
        """
        Check if the give price and time would be instantly triggered by last recent trades
        :param price: the trigger price
        :param timestamp: the timestamp to wait for
        :param trigger_above: True if waiting for an upper price
        :return: True if it would be triggered
        """
        for recent_price in self._last_recent_prices:
            trade_price = recent_price[self.PRICE_KEY]
            if timestamp <= recent_price[self.TIME_KEY] and (
                (trigger_above and price <= trade_price) or
                (not trigger_above and price >= trade_price)
            ):
                return True
        return False

    def remove_event(self, event_to_remove):
        """
        Public method that calls _remove_event()
        :param event_to_remove: the event to remove
        """
        return self._remove_event(event_to_remove)

    def _remove_and_set_event(self, event_to_set):
        """
        Set the event and remove it from event list
        :param event_to_set: the event to set
        :return: the event index removed from event list
        """
        event_to_set.set()
        return self._remove_event(event_to_set)

    def _remove_event(self, event_to_remove):
        """
        Remove the event from events list
        :param event_to_remove: the event to remove
        """
        for price_event_data in self.events:
            if event_to_remove in price_event_data:
                return self.events.remove(price_event_data)

    def _check_events(self, price, timestamp):
        """
        Check for each price, timestamp pair event if it should be triggered
        :param price: the price used to check
        :param timestamp: the timestamp used to check
        :return: the event list that match
        """
        return [
            event
            for event_price, event_timestamp, event, trigger_above in self.events
            if event_timestamp <= timestamp and
            (
                (trigger_above and event_price <= price) or
                (not trigger_above and event_price >= price)
            )
        ]


def _new_price_event(price, timestamp, trigger_above):
    """
    Create a new price event item
    :param price: the price condition
    :param timestamp: the timestamp condition
    :param trigger_above: True if waiting for an upper price
    :return: a tuple to be added into events list
    """
    return price, timestamp, asyncio.Event(), trigger_above



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import asyncio
import copy
import decimal
import types

import octobot_commons.display as commons_display
import octobot_commons.logging as logging

import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.constants as trading_constants


class AbstractStorage:
    IS_LIVE_CONSUMER = True
    USE_LIVE_CONSUMER_IN_BACKTESTING = False
    LIVE_CHANNEL = None
    IS_HISTORICAL = True
    HISTORY_TABLE = None
    FLUSH_DEBOUNCE_DURATION = 5   # avoid disc spam on multiple quick live updated

    def __init__(self, exchange_manager, plot_settings: commons_display.PlotSettings,
                 use_live_consumer_in_backtesting=None, is_historical=None):
        self.exchange_manager = exchange_manager
        self.plot_settings: commons_display.PlotSettings = plot_settings
        self.consumer = None
        self.use_live_consumer_in_backtesting = use_live_consumer_in_backtesting \
            or self.USE_LIVE_CONSUMER_IN_BACKTESTING
        self.is_historical = is_historical or self.IS_HISTORICAL
        self.enabled = True
        self._update_task = None
        self._flush_task = None
        self._to_update_auth_data_ids_buffer = set()

    def should_register_live_consumer(self):
        return self.IS_LIVE_CONSUMER and \
           (
                not self.exchange_manager.is_backtesting or
                (self.exchange_manager.is_backtesting and self.use_live_consumer_in_backtesting)
            )

    async def start(self):
        if self.should_register_live_consumer():
            await self.register_live_consumer()
        await self.on_start()

    async def on_start(self):
        """
        Called after start, implement in necessary
        """

    async def register_live_consumer(self):
        if self.LIVE_CHANNEL is None:
            raise ValueError(f"self.live_channel has to be set")
        self.consumer = await exchanges_channel.get_chan(
            self.LIVE_CHANNEL,
            self.exchange_manager.id
        ).new_consumer(
            self._live_callback
        )

    async def enable(self, enabled):
        if self.enabled != enabled:
            self.enabled = enabled
            if self.enabled:
                await self.start()
            else:
                await self.stop(clear=False)

    async def stop(self, clear=True):
        if self.consumer is not None:
            await self.consumer.stop()
        for task in (self._update_task, self._flush_task):
            if task is not None and not task.done():
                task.cancel()
        if clear:
            self.consumer = None
            self.exchange_manager = None

    async def store_history(self):
        if self.enabled:
            await self._store_history()

    async def trigger_debounced_update_auth_data(self, reset: bool):
        if self.exchange_manager.is_backtesting:
            # no interest in backtesting data
            return
        if self._update_task is not None and not self._update_task.done():
            self._update_task.cancel()
        self._update_task = asyncio.create_task(self._waiting_update_auth_data(reset))

    async def trigger_debounced_flush(self):
        if self.exchange_manager.is_backtesting:
            # flush now in backtesting
            await self.flush()
            return
        if self._flush_task is not None and not self._flush_task.done():
            self._flush_task.cancel()
        self._flush_task = asyncio.create_task(self._waiting_flush())

    async def get_history(self):
        # override if necessary
        return [
            copy.copy(document[trading_constants.STORAGE_ORIGIN_VALUE])
            for document in await self._get_db().all(self.HISTORY_TABLE)
            if trading_constants.STORAGE_ORIGIN_VALUE in document
        ]

    async def _waiting_update_auth_data(self, reset):
        try:
            await asyncio.sleep(trading_constants.AUTH_UPDATE_DEBOUNCE_DURATION)
            await self._update_auth_data(reset)
        except Exception as err:
            logging.get_logger(self.__class__.__name__).exception(err, True, f"Error when updating auth data: {err}")

    async def _waiting_flush(self):
        try:
            await asyncio.sleep(self.FLUSH_DEBOUNCE_DURATION)
            await self.flush()
        except Exception as err:
            logging.get_logger(self.__class__.__name__).exception(err, True, f"Error when flushing database: {err}")

    async def _update_auth_data(self, reset):
        pass

    async def _live_callback(self, *args, **kwargs):
        raise NotImplementedError(f"_live_callback not implemented for {self.__class__.__name__}")

    async def _store_history(self):
        raise NotImplementedError(f"_store_history not implemented for {self.__class__.__name__}")

    def _get_db(self, *args):
        raise NotImplementedError(f"_get_db not implemented for {self.__class__.__name__}")

    async def clear_history(self, flush=True):
        await self.clear_database_history(self._get_db(), flush=flush)

    async def flush(self):
        await self._get_db().flush()

    @classmethod
    async def clear_database_history(cls, database, flush=True):
        if cls.HISTORY_TABLE is None:
            raise NotImplementedError(f"{cls.__name__}.HISTORY_TABLE has to be set")
        await database.delete(cls.HISTORY_TABLE, None)
        if flush:
            await database.flush()

    @classmethod
    def sanitize_for_storage(cls, element: dict) -> dict:
        sanitized = copy.copy(element)
        for key, val in element.items():
            if isinstance(val, decimal.Decimal):
                sanitized[key] = float(val)
            elif isinstance(val, dict):
                sanitized[key] = cls.sanitize_for_storage(val)
            elif isinstance(val, types.FunctionType):
                raise ValueError(f"{val.__name__} is a function, it can't be serialized")
        return sanitized

    @staticmethod
    def hard_reset_and_retry_if_necessary(fn):
        """
        Will retry the given function if a database hard reset error is raised
        Warning: when it happens, will also completely reset the database
        """
        async def wrapper(*args, **kwargs):
            try:
                return await fn(*args, **kwargs)
            except Exception as err:
                database = args[0]._get_db()    # pylint: disable=protected-access
                if database.is_hard_reset_error(err):
                    logging.get_logger(args[0].__class__.__name__).warning(
                        f"Resetting database due to [{err}] error"
                    )
                    await database.hard_reset()
                    return await fn(*args, **kwargs)
                raise
        return wrapper



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from octobot_trading.enums import TraderOrderType
from octobot_commons.asyncio_tools import wait_asyncio_next_cycle

import octobot_trading.constants as trading_constants
from tests.personal_data import DEFAULT_MARKET_QUANTITY, DEFAULT_ORDER_SYMBOL
from tests.test_utils.random_numbers import decimal_random_price, decimal_random_quantity, \
    decimal_random_recent_trade

from tests import event_loop
from tests.exchanges import simulated_trader, simulated_exchange_manager
from tests.personal_data.orders import buy_limit_order

pytestmark = pytest.mark.asyncio


async def test_buy_limit_order_trigger(buy_limit_order):
    order_price = decimal_random_price()
    buy_limit_order.update(
        price=order_price,
        quantity=decimal_random_quantity(max_value=DEFAULT_MARKET_QUANTITY / order_price),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=TraderOrderType.BUY_LIMIT,
    )
    buy_limit_order.exchange_manager.is_backtesting = True  # force update_order_status
    await buy_limit_order.initialize()
    await buy_limit_order.exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(
        buy_limit_order
    )
    price_events_manager = buy_limit_order.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(
        DEFAULT_ORDER_SYMBOL).price_events_manager
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=decimal_random_price(min_value=order_price + trading_constants.ONE),
                                     timestamp=buy_limit_order.timestamp)])
    await wait_asyncio_next_cycle()
    assert not buy_limit_order.is_filled()
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=order_price,
                                     timestamp=buy_limit_order.timestamp - 1)])
    await wait_asyncio_next_cycle()
    assert not buy_limit_order.is_filled()
    price_events_manager.handle_recent_trades([decimal_random_recent_trade(price=order_price,
                                                                           timestamp=buy_limit_order.timestamp)])

    await wait_asyncio_next_cycle()
    assert buy_limit_order.is_filled()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging

import octobot_trading.enums as enums
import octobot_trading.constants as constants


def update_contracts_from_positions(exchange_manager, positions) -> bool:
    updated = False
    for position in positions:
        pair = position.get(enums.ExchangeConstantsPositionColumns.SYMBOL.value, None)
        if pair and pair not in exchange_manager.exchange.pair_contracts:
            contract = exchange_manager.exchange.create_pair_contract(
                pair=pair,
                current_leverage=position.get(
                    enums.ExchangeConstantsPositionColumns.LEVERAGE.value, constants.ZERO),
                contract_size=position.get(
                    enums.ExchangeConstantsPositionColumns.CONTRACT_SIZE.value, constants.DEFAULT_SYMBOL_CONTRACT_SIZE),
                margin_type=position.get(enums.ExchangeConstantsPositionColumns.MARGIN_TYPE.value, None),
                contract_type=position.get(enums.ExchangeConstantsPositionColumns.CONTRACT_TYPE.value, None),
                position_mode=position.get(enums.ExchangeConstantsPositionColumns.POSITION_MODE.value, None),
                maintenance_margin_rate=position.get(
                    enums.ExchangeConstantsPositionColumns.MAINTENANCE_MARGIN_RATE.value,
                    constants.DEFAULT_SYMBOL_MAINTENANCE_MARGIN_RATE)
            )
            # contracts are updated from positions are fully initialized
            if contract.is_handled_contract():
                updated = True
            else:
                message = f"Unhandled contract {contract}. This contract can't be traded"
                if pair in exchange_manager.exchange_config.traded_symbol_pairs:
                    # inform user that the contract can't be used
                    _get_logger().error(message)
                else:
                    # no need to inform as the contract is not requested
                    _get_logger().debug(message)
    return updated


def _get_logger():
    return logging.get_logger("contract_factory")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import pytest
from octobot_commons.constants import PORTFOLIO_TOTAL, PORTFOLIO_AVAILABLE

from octobot_trading.enums import TraderOrderType
from octobot_trading.personal_data.orders.types.limit.sell_limit_order import SellLimitOrder
from octobot_trading.personal_data.orders.types.market.buy_market_order import BuyMarketOrder
from octobot_trading.personal_data.portfolios.types.spot_portfolio import SpotPortfolio
from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_initial_portfolio(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    assert isinstance(portfolio_manager.portfolio, SpotPortfolio)


async def test_update_portfolio_data_from_order(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    pass


async def test_update_portfolio_available_from_order(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # Test buy order
    market_buy = BuyMarketOrder(trader)
    market_buy.update(order_type=TraderOrderType.BUY_MARKET,
                      symbol="BTC/USDT",
                      current_price=70,
                      quantity=10,
                      price=70)

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('300')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # test buy order canceled --> return to init state and the update_portfolio will sync TOTAL with AVAILABLE
    portfolio_manager.portfolio.update_portfolio_available(market_buy, False)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # Test sell order
    limit_sell = SellLimitOrder(trader)
    limit_sell.update(order_type=TraderOrderType.SELL_LIMIT,
                      symbol="BTC/USDT",
                      current_price=60,
                      quantity=8,
                      price=60)

    # test sell order creation
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('2')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # test sell order canceled --> return to init state and the update_portfolio will sync TOTAL with AVAILABLE
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, False)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
from ccxt.async_support import kraken

from octobot_commons.enums import TimeFrames, PriceIndexes
from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc, \
    ExchangeConstantsOrderBookInfoColumns as Ecobic, ExchangeConstantsOrderColumns as Ecoc, \
    ExchangeConstantsTickersColumns as Ectc
from tests_additional.real_exchanges.real_exchange_tester import RealExchangeTester
# required to catch async loop context exceptions
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestKrakenRealExchangeTester(RealExchangeTester):
    EXCHANGE_NAME = "kraken"
    SYMBOL = "BTC/USDT"
    SYMBOL_2 = "ETH/BTC"
    SYMBOL_3 = "XRP/BTC"

    async def test_time_frames(self):
        time_frames = await self.time_frames()
        assert all(time_frame in time_frames for time_frame in (
            TimeFrames.ONE_MINUTE.value,
            TimeFrames.FIVE_MINUTES.value,
            TimeFrames.FIFTEEN_MINUTES.value,
            TimeFrames.THIRTY_MINUTES.value,
            TimeFrames.ONE_HOUR.value,
            TimeFrames.FOUR_HOURS.value,
            TimeFrames.ONE_DAY.value,
            TimeFrames.ONE_WEEK.value
        ))

    async def test_active_symbols(self):
        await self.inner_test_active_symbols(700, 700)

    async def test_get_market_status(self):
        for market_status in await self.get_market_statuses():
            self.ensure_required_market_status_values(market_status)
            # on this exchange, precision is a decimal instead of a number of digits
            assert 0 < market_status[Ecmsc.PRECISION.value][
                Ecmsc.PRECISION_AMOUNT.value] <= 1  # to be fixed in this exchange tentacle
            assert 0 < market_status[Ecmsc.PRECISION.value][
                Ecmsc.PRECISION_PRICE.value] <= 1  # to be fixed in this exchange tentacle
            assert all(elem in market_status[Ecmsc.LIMITS.value]
                       for elem in (Ecmsc.LIMITS_AMOUNT.value,
                                    Ecmsc.LIMITS_PRICE.value,
                                    Ecmsc.LIMITS_COST.value))
            self.check_market_status_limits(market_status,
                                            normal_cost_min=1e-07,
                                            low_cost_min=1e-08,
                                            expect_invalid_price_limit_values=False,
                                            enable_price_and_cost_comparison=False)

    async def test_get_symbol_prices(self):
        # without limit
        symbol_prices = await self.get_symbol_prices()
        assert len(symbol_prices) == 720
        # check candles order (oldest first)
        self.ensure_elements_order(symbol_prices, PriceIndexes.IND_PRICE_TIME.value)
        # check last candle is the current candle
        assert symbol_prices[-1][PriceIndexes.IND_PRICE_TIME.value] >= self.get_time() - self.get_allowed_time_delta()

        # candle limit is not supported, replaced by (await self.get_symbol_prices())[-limit:] in kraken tentacle
        # try with candles limit (used in candled updater)
        symbol_prices = (await self.get_symbol_prices())[-200:]
        assert len(symbol_prices) == 200
        # check candles order (oldest first)
        self.ensure_elements_order(symbol_prices, PriceIndexes.IND_PRICE_TIME.value)
        # check last candle is the current candle
        assert symbol_prices[-1][PriceIndexes.IND_PRICE_TIME.value] >= self.get_time() - self.get_allowed_time_delta()

    async def test_get_historical_symbol_prices(self):
        # try with since and limit (used in data collector)
        for limit in (50, None):
            symbol_prices = await self.get_symbol_prices(since=self.CANDLE_SINCE, limit=limit)
            if limit:
                assert len(symbol_prices) == limit
            else:
                assert len(symbol_prices) > 5
            # check candles order (oldest first)
            self.ensure_elements_order(symbol_prices, PriceIndexes.IND_PRICE_TIME.value)
            # check that fetched candles are historical candles
            max_candle_time = self.get_time_after_time_frames(self.CANDLE_SINCE_SEC, len(symbol_prices))
            assert max_candle_time <= self.get_time()
            with pytest.raises(AssertionError):  # not supported
                for candle in symbol_prices:
                    assert self.CANDLE_SINCE_SEC <= candle[PriceIndexes.IND_PRICE_TIME.value] <= max_candle_time

    async def test_get_historical_ohlcv(self):
        # not supported
        ohlcv = await self.get_historical_ohlcv()
        assert 0 < len(ohlcv) < 100

    async def test_get_kline_price(self):
        # kline_price = await self.get_kline_price()
        client = kraken()
        await client.fetch_markets()
        kline_price = [(await client.fetch_ohlcv(TestKrakenRealExchangeTester.SYMBOL, TimeFrames.ONE_HOUR.value))[-1]]

        assert len(kline_price) == 1
        assert len(kline_price[0]) == 6
        kline_start_time = kline_price[0][PriceIndexes.IND_PRICE_TIME.value]
        # assert kline is the current candle
        assert kline_start_time >= self.get_time() - self.get_allowed_time_delta()
        await client.close()

    async def test_get_order_book(self):
        # We should prefer using fetch_l2_order_book (from https://github.com/ccxt/ccxt/issues/8135)
        order_book = await self.get_order_book()
        assert 0 < order_book[Ecobic.TIMESTAMP.value] < self._get_ref_order_book_timestamp()
        assert len(order_book[Ecobic.ASKS.value]) == 5
        # [price, amount, time] instead of [price, amount]
        assert len(order_book[Ecobic.ASKS.value][0]) == 3
        assert len(order_book[Ecobic.BIDS.value]) == 5
        # [price, amount, time] instead of [price, amount]
        assert len(order_book[Ecobic.BIDS.value][0]) == 3
        
    async def test_get_order_books(self):
        # implement if necessary
        pass

    async def test_get_recent_trades(self):
        # https://github.com/ccxt/ccxt/issues/5698
        recent_trades = await self.get_recent_trades(limit=1000)
        assert len(recent_trades) == 1000
        # check trades order (oldest first)
        self.ensure_elements_order(recent_trades, Ecoc.TIMESTAMP.value)

    async def test_get_price_ticker(self):
        ticker = await self.get_price_ticker()
        self._check_ticker(ticker, self.SYMBOL, check_content=True)

    async def test_get_all_currencies_price_ticker(self):
        # get_all_currencies_price_ticker kraken fetchTickers() requires a symbols argument, an array of symbols
        tickers = await self.get_all_currencies_price_ticker(symbols=[self.SYMBOL, self.SYMBOL_2])
        for symbol, ticker in tickers.items():
            self._check_ticker(ticker, symbol)

    @staticmethod
    def _check_ticker(ticker, symbol, check_content=False):
        assert ticker[Ectc.SYMBOL.value] == symbol
        assert all(key in ticker for key in (
            Ectc.HIGH.value,
            Ectc.LOW.value,
            Ectc.BID.value,
            Ectc.BID_VOLUME.value,
            Ectc.ASK.value,
            Ectc.ASK_VOLUME.value,
            Ectc.OPEN.value,
            Ectc.CLOSE.value,
            Ectc.LAST.value,
            Ectc.PREVIOUS_CLOSE.value
        ))
        if check_content:
            assert ticker[Ectc.HIGH.value]
            assert ticker[Ectc.LOW.value]
            assert ticker[Ectc.BID.value]
            assert ticker[Ectc.BID_VOLUME.value] is None
            assert ticker[Ectc.ASK.value]
            assert ticker[Ectc.ASK_VOLUME.value] is None
            assert ticker[Ectc.OPEN.value]
            assert ticker[Ectc.CLOSE.value]
            assert ticker[Ectc.LAST.value]
            assert ticker[Ectc.PREVIOUS_CLOSE.value] is None
            assert ticker[Ectc.BASE_VOLUME.value]
            assert ticker[Ectc.TIMESTAMP.value] is None  # will trigger an 'Ignored incomplete ticker'
            RealExchangeTester.check_ticker_typing(ticker, check_timestamp=False)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging

import octobot_trading.personal_data as personal_data
import octobot_trading.enums as enums
import octobot_trading.constants as constants


def create_order_from_raw(trader, raw_order):
    _, order_type = personal_data.parse_order_type(raw_order)
    return create_order_from_type(trader, order_type)


def create_order_instance_from_raw(
    trader, raw_order, force_open_or_pending_creation=False, has_just_been_created=False
):
    try:
        order = create_order_from_raw(trader, raw_order)
        order.update_from_raw(raw_order)
        if has_just_been_created:
            order.register_broker_applied_if_enabled()
        if force_open_or_pending_creation \
                and order.status not in (enums.OrderStatus.OPEN, enums.OrderStatus.PENDING_CREATION):
            order.status = enums.OrderStatus.OPEN
        return order
    except Exception as err:
        # log unparsable order to fix it
        logging.get_logger(__name__).exception(
            err, True, f"Unexpected {err} ({err.__class__.__name__}) error when parsing row order {raw_order}"
        )
        raise


def create_order_from_type(trader, order_type, side=None):
    if side is None:
        return personal_data.TraderOrderTypeClasses[order_type](trader)
    return personal_data.TraderOrderTypeClasses[order_type](trader, side=side)


def create_order_instance(
    trader,
    order_type,
    symbol,
    current_price,
    quantity,
    price=constants.ZERO,
    stop_price=constants.ZERO,
    status=enums.OrderStatus.OPEN,
    order_id=None,
    exchange_order_id=None,
    filled_price=constants.ZERO,
    average_price=constants.ZERO,
    quantity_filled=constants.ZERO,
    total_cost=constants.ZERO,
    timestamp=0,
    side=None,
    fees_currency_side=None,
    group=None,
    tag=None,
    reduce_only=None,
    quantity_currency=None,
    close_position=False,
    exchange_creation_params=None,
    associated_entry_id=None
):
    order = create_order_from_type(trader=trader,
                                   order_type=order_type,
                                   side=side)
    order.update(
        order_type=order_type,
        symbol=symbol,
        current_price=current_price,
        quantity=quantity,
        price=price,
        stop_price=stop_price,
        order_id=trader.parse_order_id(order_id),
        exchange_order_id=exchange_order_id,
        timestamp=timestamp,
        status=status,
        filled_price=filled_price,
        average_price=average_price,
        quantity_filled=quantity_filled,
        fee=None,
        total_cost=total_cost,
        fees_currency_side=fees_currency_side,
        group=group,
        tag=tag,
        reduce_only=reduce_only,
        quantity_currency=quantity_currency,
        close_position=close_position,
        exchange_creation_params=exchange_creation_params,
        associated_entry_id=associated_entry_id
    )
    return order


def create_order_from_dict(trader, order_dict):
    """
    :param trader: the trader to associate the order to
    :param order_dict: a dict formatted as from order.to_dict()
    :return: the created order instance
    """
    _, order_type = personal_data.parse_order_type(order_dict)
    return create_order_instance(
        trader,
        order_type,
        order_dict[enums.ExchangeConstantsOrderColumns.SYMBOL.value],
        constants.ZERO,
        order_dict[enums.ExchangeConstantsOrderColumns.AMOUNT.value],
        price=order_dict[enums.ExchangeConstantsOrderColumns.PRICE.value],
        status=enums.OrderStatus(order_dict[enums.ExchangeConstantsOrderColumns.STATUS.value]),
        order_id=order_dict[enums.ExchangeConstantsOrderColumns.ID.value],
        quantity_filled=order_dict[enums.ExchangeConstantsOrderColumns.FILLED.value],
        timestamp=order_dict[enums.ExchangeConstantsOrderColumns.TIMESTAMP.value],
        side=enums.TradeOrderSide(order_dict[enums.ExchangeConstantsOrderColumns.SIDE.value]),
        tag=order_dict[enums.ExchangeConstantsOrderColumns.TAG.value],
        reduce_only=order_dict[enums.ExchangeConstantsOrderColumns.REDUCE_ONLY.value],
    )


async def create_order_from_order_storage_details(order_storage_details, exchange_manager, pending_groups: dict):
    order = create_order_from_dict(
        exchange_manager.trader,
        order_storage_details[constants.STORAGE_ORIGIN_VALUE]
    )
    order.update_from_storage_order_details(order_storage_details)
    await personal_data.create_orders_storage_related_elements(
        order, order_storage_details, exchange_manager, pending_groups
    )
    return order


async def restore_chained_orders_from_storage_order_details(order, order_details, exchange_manager, pending_groups):
    chained_orders = order_details.get(enums.StoredOrdersAttr.CHAINED_ORDERS.value, None)
    if chained_orders:
        for chained_order in chained_orders:
            chained_order_inst = await create_order_from_order_storage_details(
                chained_order, exchange_manager, pending_groups
            )
            await chained_order_inst.set_as_chained_order(
                order,
                chained_order.get(enums.StoredOrdersAttr.HAS_BEEN_BUNDLED.value, False),
                chained_order.get(enums.StoredOrdersAttr.EXCHANGE_CREATION_PARAMS.value, {}),
                chained_order.get(enums.StoredOrdersAttr.UPDATE_WITH_TRIGGERING_ORDER_FEES.value, False),
                **chained_order.get(enums.StoredOrdersAttr.TRADER_CREATION_KWARGS.value, {}),
            )
            order.add_chained_order(chained_order_inst)
            logging.get_logger(order.get_logger_name()).debug(f"Restored chained order: {chained_order_inst}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.exchange_data as exchange_data


def is_inverse_future_contract(contract_type):
    return exchange_data.FutureContract(None, None, contract_type).is_inverse_contract()


def is_perpetual_future_contract(contract_type):
    return exchange_data.FutureContract(None, None, contract_type).is_perpetual_contract()


def get_pair_contracts(exchange_manager) -> dict:
    return exchange_manager.exchange.pair_contracts


def is_handled_contract(contract) -> bool:
    return contract.is_handled_contract()



#  Drakkar-Software OctoBot-Private-Tentacles
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal
import time
import typing

import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.enums as common_enums
import octobot_commons.constants as commons_constants
import octobot_commons.logging as logging
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.api as trading_api
import octobot_trading.exchanges as exchanges
import octobot_tentacles_manager.api as tentacles_manager_api

import octobot_trading.personal_data as personal_data

import octobot_trading.util.test_tools.exchange_data as exchange_data_import


BASE_TIMEOUT = 10


async def create_test_exchange_manager(
        config: dict,
        exchange_name: str,
        is_spot_only: bool = True,
        is_margin: bool = False,
        is_future: bool = False,
        rest_only: bool = False,
        is_real: bool = True,
        is_sandboxed: bool = False,
        ignore_exchange_config: bool = True) -> exchanges.ExchangeManager:
    if ignore_exchange_config:
        # enable exchange name in config
        config[commons_constants.CONFIG_EXCHANGES][exchange_name] = {commons_constants.CONFIG_ENABLED_OPTION: True}

    builder = exchanges.create_exchange_builder_instance(config, exchange_name)
    builder.disable_trading_mode()
    builder.use_tentacles_setup_config(tentacles_manager_api.get_tentacles_setup_config())
    if is_spot_only:
        builder.is_spot_only()
    if is_margin:
        builder.is_margin()
    if is_future:
        builder.is_future()
    if rest_only:
        builder.is_rest_only()
    if is_sandboxed:
        builder.is_sandboxed(is_sandboxed)
    if is_real:
        builder.is_real()
    else:
        builder.is_simulated()
    await builder.build()
    return builder.exchange_manager


async def stop_test_exchange_manager(exchange_manager_instance: exchanges.ExchangeManager):
    trading_api.cancel_ccxt_throttle_task()
    await exchange_manager_instance.stop()
    # let updaters gracefully shutdown
    await asyncio_tools.wait_asyncio_next_cycle()


async def add_symbols_details(
    exchange_manager, symbols: list, time_frame: str, exchange_data: exchange_data_import.ExchangeData,
    history_size=1, start_time=0, end_time=0, close_price_only=False,
    include_latest_candle=True, reload_markets=False,
    market_filter: typing.Union[None, typing.Callable[[dict], bool]] = None
) -> exchange_data_import.ExchangeData:
    parsed_tf = common_enums.TimeFrames(time_frame)

    async def _update_ohlcv(symbol):
        if start_time == 0:
            ohlcvs = await exchange_manager.exchange.get_symbol_prices(symbol, parsed_tf, limit=history_size)
        else:
            ohlcvs = []
            async for ohlcv in exchanges.get_historical_ohlcv(
                exchange_manager, symbol, parsed_tf, start_time * 1000, (end_time or time.time()) * 1000,
                request_retry_timeout=BASE_TIMEOUT
            ):
                ohlcvs.extend(ohlcv)
        if not include_latest_candle:
            ohlcvs = ohlcvs[:-1]
        details = exchange_data_import.MarketDetails(
            symbol=symbol,
            time_frame=time_frame,
            close=[ohlcv[common_enums.PriceIndexes.IND_PRICE_CLOSE.value] for ohlcv in ohlcvs],
            open=[ohlcv[common_enums.PriceIndexes.IND_PRICE_OPEN.value] for ohlcv in ohlcvs] if not close_price_only else [],
            high=[ohlcv[common_enums.PriceIndexes.IND_PRICE_HIGH.value] for ohlcv in ohlcvs] if not close_price_only else [],
            low=[ohlcv[common_enums.PriceIndexes.IND_PRICE_LOW.value] for ohlcv in ohlcvs] if not close_price_only else [],
            volume=[ohlcv[common_enums.PriceIndexes.IND_PRICE_VOL.value] for ohlcv in ohlcvs] if not close_price_only else [],
            time=[ohlcv[common_enums.PriceIndexes.IND_PRICE_TIME.value] for ohlcv in ohlcvs],
        )
        exchange_data.markets.append(details)

    await ensure_symbol_markets(exchange_manager, reload=reload_markets, market_filter=market_filter)
    await asyncio.gather(*(_update_ohlcv(symbol) for symbol in symbols))
    return exchange_data


async def ensure_symbol_markets(
    exchange_manager,
    reload=False,
    market_filter: typing.Union[None, typing.Callable[[dict], bool]] = None
):
    if reload or not exchange_manager.exchange.connector.has_markets():
        await exchange_manager.exchange.connector.load_symbol_markets(reload=reload, market_filter=market_filter)


async def get_portfolio(exchange_manager, as_float=False, clear_empty=True) -> dict:
    balance = await exchange_manager.exchange.get_balance()
    # filter out 0 values
    return {
        asset: {key: float(val) if as_float else val for key, val in values.items()}  # use float for values
        for asset, values in balance.items()
        if not clear_empty or (clear_empty and any(value for value in values.values()))
    }


async def get_open_orders(
    exchange_manager,
    exchange_data: exchange_data_import.ExchangeData,
    symbols: list = None
) -> list:
    open_orders = []

    async def _get_orders(symbol):
        orders = await exchange_manager.exchange.get_open_orders(symbol=symbol)
        open_orders.extend(
            personal_data.create_order_instance_from_raw(
                exchange_manager.trader, order, force_open_or_pending_creation=True
            ).to_dict()
            for order in orders
        )

    symbols = symbols or [market.symbol for market in exchange_data.markets]
    await asyncio.gather(*(_get_orders(symbol) for symbol in symbols))
    return open_orders


async def get_order(
    exchange_manager,
    exchange_order_id: str,
    symbol: str,
) -> typing.Optional[dict]:
    return await exchange_manager.exchange.get_order(exchange_order_id, symbol=symbol)


async def get_cancelled_orders(
    exchange_manager,
    exchange_data: exchange_data_import.ExchangeData,
    symbols: list = None
) -> list:
    cancelled_orders = []

    async def _get_orders(symbol):
        orders = await exchange_manager.exchange.get_cancelled_orders(symbol=symbol)
        cancelled_orders.extend(
            personal_data.create_order_instance_from_raw(
                exchange_manager.trader, order, force_open_or_pending_creation=False
            ).to_dict()
            for order in orders
        )

    symbols = symbols or [market.symbol for market in exchange_data.markets]
    await asyncio.gather(*(_get_orders(symbol) for symbol in symbols))
    return cancelled_orders


async def get_trades(
    exchange_manager,
    exchange_data: exchange_data_import.ExchangeData,
    symbols: list = None
) -> list:
    trades = []

    async def _get_trades(symbol):
        row_trades = await exchange_manager.exchange.get_my_recent_trades(symbol=symbol)
        trades.extend(
            personal_data.create_trade_instance_from_raw(exchange_manager.trader, raw_trade).to_dict()
            for raw_trade in row_trades
        )

    symbols = symbols or [market.symbol for market in exchange_data.markets]
    await asyncio.gather(*(_get_trades(symbol) for symbol in symbols))
    return trades


async def create_orders(
    exchange_manager,
    exchange_data: exchange_data_import.ExchangeData,
    orders: list,
    order_creation_timeout: float
) -> list:
    async def _create_order(order_dict) -> personal_data.Order:
        symbol = order_dict[enums.ExchangeConstantsOrderColumns.SYMBOL.value]
        side, order_type = personal_data.parse_order_type(order_dict)
        created_order = await exchange_manager.exchange.create_order(
            order_type,
            symbol,
            decimal.Decimal(str(order_dict[enums.ExchangeConstantsOrderColumns.AMOUNT.value])),
            price=decimal.Decimal(str(order_dict[enums.ExchangeConstantsOrderColumns.PRICE.value])),
            side=side,
            current_price=decimal.Decimal(str(exchange_data.get_price(symbol))),
            reduce_only=order_dict[enums.ExchangeConstantsOrderColumns.REDUCE_ONLY.value],
        )
        # is private, to use in tests context only
        order = personal_data.create_order_instance_from_raw(
            exchange_manager.trader, created_order, force_open_or_pending_creation=True
        ) if created_order else None
        if not order:
            return order
        if order.status is enums.OrderStatus.PENDING_CREATION and order_creation_timeout > 0:
            try:
                return await wait_for_other_status(order, order_creation_timeout)
            except TimeoutError as err:
                logging.get_logger(order.get_logger_name()).error(f"Created order can't be fetched: {err}")
                return None
        return order

    return await asyncio.gather(*(_create_order(order_dict) for order_dict in orders))


async def wait_for_other_status(order: personal_data.Order, timeout) -> personal_data.Order:
    t0 = time.time()
    iterations = 0
    origin_status = order.status.value
    while time.time() - t0 < timeout:
        raw_order = await order.exchange_manager.exchange.get_order(order.exchange_order_id, order.symbol)
        iterations += 1
        if raw_order is not None and raw_order[enums.ExchangeConstantsOrderColumns.STATUS.value] != origin_status:
            logging.get_logger(order.get_logger_name()).info(
                f"Order fetched with status different from {origin_status} after {iterations} "
                f"iterations and {round(time.time() - t0)}s"
            )
            return personal_data.create_order_instance_from_raw(
                order.trader, raw_order, force_open_or_pending_creation=False
            )
        if time.time() - t0 + constants.CREATED_ORDER_FORCED_UPDATE_PERIOD >= timeout:
            break
        await asyncio.sleep(constants.CREATED_ORDER_FORCED_UPDATE_PERIOD)
    raise TimeoutError(f"Order was not found with another status than {origin_status} within {timeout} seconds")
